{----------------------------------------------------------------------------
|
| Library: Spider Containers for Object Pascal
|
| Module: ObjBase.Pas
|
| Description: Abstract classes for generic containers and general type
|              definitions.
|
| Copyright (c) 1996,97 Michel Brazeau, Interval Software
|
| History: Mar  , 1996  Version 1.0.
|          Oct 31, 1996. MB, Made the Exception types public by moving them to
|                        the Interface section.
|          Apr 28, 1997. MB added exceptions for object streaming
|          May 10, 1997. MB added Ownership property.
|          May 11, 1997. MB moved exceptions to ObjType unit.
|          May 12, 1997. MB add CallBackData parameter to TProcIterator,
|                        TMethodIterator, ForEachCallProc, ForEachCallMethod
|          Jan 17, 1998. MB Resource strings (in Spider16.Res or Spider32.Res
|                        files) are no longer used for exception messages to
|                        avoid conflicts with other third party libraries. The
|                        strings are created in the ObjType unit, at may be
|                        modified at run-time
|
|---------------------------------------------------------------------------}
unit ObjBase;

{$I Contain.Inc}

interface

uses
    SysUtils, { Exception }
    ObjPerst, { TObjectReader, TObjectWriter }
    ObjCmp;   { for TCompareFunc }
       
type

{ iterator routines for file level procedures and methods }
TProcIterator   = procedure( const Obj          : TObject;
                             const CallBackData : LongInt );

TMethodIterator = procedure( const Obj          : TObject;
                             const CallBackData : LongInt ) of object;

type

{ base class for all containers }
TContainer = class(TObject)
protected
    FObjectType   : TClass;       { type of element allowed in container }

    FCompareFunc  : TCompareFunc; { comparison function }

    FSize         : LongInt;      { number of elements in container }

    FOwnership    : Boolean;      { if True contained objects are disposed }

    { verifies to make sure Obj is of Type FObjectType. If not, the
      EInvalidObjectType exception is raised }
    procedure CheckType(const Obj : TObject);

    { GetSize and QueryEmpty may overriden for derived containers that
      use an embedded container }
    function GetSize    : LongInt; virtual;
    function QueryEmpty : Boolean; virtual;

    function GetOwnership : Boolean; virtual;
    procedure SetOwnership( const InOwnership : Boolean ); virtual;

    { releases an object from the container. If Ownership if True, the object
      is freed, otherwise Obj is not freed }
    procedure ReleaseObj( const Obj : TObject );

public
    constructor Create(const InObjectType : TClass);

    { destroys calls Clear to remove and free every element in the container }
    destructor Destroy; override;

    { number of elements in container, read only }
    property Size : LongInt read GetSize;

    { True is container is emtpy, read only }
    property IsEmpty : Boolean read QueryEmpty;

    { for each item in the container, call ProcIterator. ProcIterator is
      a file level procedure. }
    procedure ForEachCallProc( const ProcIterator : TProcIterator;
                               const CallBackData : LongInt ); virtual; abstract;

    { for each item in the container, call MethodIterator. MethodIterator is
      a method of a class. }
    procedure ForEachCallMethod( const MethodIterator : TMethodIterator;
                                 const CallBackData   : LongInt ); virtual; abstract;

    { clear all the items in the container. Free is called for each item }
    procedure Clear; virtual; abstract;

   { if True contained objects are disposed. Default is True }
    property Ownership : Boolean read GetOwnership
                                 write SetOwnership;

end;

{ Abstract node class used in node based containers. Derived containers
  must augment the data members for the proper container type. }
TNode = class(TObject)
public
    Obj      : TObject;

    constructor Create;

    destructor Destroy; override; { calls Obj.Free }
end;

type

{ base class for node based (ie lists, trees) containers }
TNodeContainer = class(TContainer)
protected
    function  GetCurrentObj : TObject; virtual;

    procedure SetCurrentObj(const Obj : TObject); virtual;

    { returns the Nth (1 based) object in the container. If if N > Size then
      EInvalidIndex is raised }
    function GetNthObj(const Index : LongInt) : TObject; virtual; abstract;

public
    { CurrentObj is used to examine the current object positioned in the
      node structure. If the container is Empty, when reading or writing to
      the current object the EContainerIsEmpty exception is raised.
      By default, this is a read only property to prevent changing the
      current object in an ordered container }
    property CurrentObj : TObject read  GetCurrentObj;

    constructor Create( const InObjectType  : TClass;
                        const InCompareFunc : TCompareFunc );

    { destroys calls Clear to remove and free every object in the container }
    destructor Destroy; override;

    { inserts a node at the proper position according to the compare
      function }
    procedure Insert(const Obj : TObject); virtual; abstract;

    { Deletes (removes and Frees) the N'th matching Occur'ence of Obj.
      Returns True if the object is found and deleted. Positioning after
      deletion is dependent of the derived container }
    function Delete( const Obj   : TObject;
                     const Occur : LongInt
                    ) : Boolean; virtual; abstract;

    { Detaches the current object. The object is removed from the container,
      and returned to the calling block. It is then the responsibility
      of the programmer to call the Free method of the object. May raise
      EContainerIsEmpty. }
    function DetachCurrent : TObject; virtual; abstract;

    { Same as DetachCurrent, but the object is Destroyed and not returned.
      By default, a call to DetachCurrent.Free is made }
    procedure DeleteCurrent; virtual;

    { searches for the N'th matching Occur'ence of Obj, and if found, the
      current node is set to it. If not found, the positioning after the
      search is dependent on the derived container }
    function Search( const Obj   : TObject;
                     const Occur : LongInt
                    ) : Boolean; virtual; abstract;

    { property that allows to access an object given a numeric index,
      1 <= Index <= Size. This is the default property, so it does not need
      to be stated explicitely, when accessing it, ie Container[1] is
      equivalent to Container.At[1]. If Index is valid, the corresponding
      Object is returned, and the current position is set to this object.
      If Index is out of range EInvalidIndex is raised. This is a read only
      property. Objects should only be added using the Insert method. Note
      that since node based containers do not use arrays for implemention,
      this method of access will be slow in large containers. It is
      preferable to use the Search and GotoXXXX methods in derived
      containers }
    property At[const Index : LongInt] : TObject read GetNthObj; default;

    { positions the current object at the Nth (1 based) position.  If
      N > Size then EInvalidIndex is raised. This is similar to the At
      property, but not any object is returned. Uses GetNthObj for its
      implementation }
    procedure GotoNth(const Index : LongInt);

    { GotoFirst, GotoLast return False when the container is emtpy.
      EContainerIsEmpty is not raised. }
    function GotoFirst : Boolean; virtual; abstract;
    function GotoLast  : Boolean; virtual; abstract;

    { returns False if container is empty or last node is current }
    function GotoNext : Boolean; virtual; abstract;

    { returns False if container is empty or first node is current }
    function GotoPrev : Boolean; virtual; abstract;

    { methods to Set and Goto a marker within the node structure. Only one
      marker may be set at a time. Invoking SetMarker overwrites the
      current marker set. }
    procedure SetMarker;
    procedure GotoMarker;

    { returns True if Current node is on the Marker, Current = Marker }
    function IsOnMarker : Boolean;

protected

    Marker   : TNode;       { marker used by SetMarker, GotoMarker }
    Current  : TNode;       { pointer to current node }

end;

{ stream registration routines }
procedure ReadContainer( const Obj    : TObject;
                         const Reader : TObjectReader ); far;

procedure WriteContainer( const Obj    : TObject;
                          const Writer : TObjectWriter ); far;

procedure ReadNodeContainer( const Obj    : TObject;
                             const Reader : TObjectReader ); far;

procedure WriteNodeContainer( const Obj    : TObject;
                              const Writer : TObjectWriter ); far;


{--------------------------------------------------------------------------}

implementation

uses
    {$ifdef __Evaluation}
    WinTypes,  { HWnd }
    WinProcs,  { FindWindow }
    Dialogs,   { MessageDlg }
    {$endif}

    ObjType;   { RaiseXXX }



{ TNode implementation ----------------------------------------------------}

constructor TNode.Create;
begin
    inherited Create;
end;

{--------------------------------------------------------------------------}

destructor TNode.Destroy;
begin
    Obj.Free;

    inherited Destroy;
end;


{ TContainer implementation -------------------------------------------------- }

constructor TContainer.Create(const InObjectType : TClass);
begin
    inherited Create;

    FObjectType := InObjectType;
    FSize       := 0;
    FOwnership  := True;
end;

{ ------------------------------------------------------------------------}

destructor TContainer.Destroy;
begin
    Clear;

    inherited Destroy;
end;

{ ------------------------------------------------------------------------}

function TContainer.QueryEmpty : Boolean;
begin
    Result := (FSize = 0);
end;

{ ------------------------------------------------------------------------}

function TContainer.GetSize : LongInt;
begin
    Result := FSize;
end;

{-------------------------------------------------------------------------}

procedure TContainer.CheckType(const Obj : TObject);
begin
    if not (Obj is FObjectType) then
        RaiseInvalidObjectType;
end;

{-------------------------------------------------------------------------}

function TContainer.GetOwnership : Boolean;
begin
    Result := FOwnership;
end;

{-------------------------------------------------------------------------}

procedure TContainer.SetOwnership( const InOwnership : Boolean );
begin
    FOwnership := InOwnership;
end;

{-------------------------------------------------------------------------}

procedure TContainer.ReleaseObj( const Obj : TObject );
begin
    if Ownership then
        Obj.Free;
end;

{ TNodeContainer implementation ----------------------------------------- }

constructor TNodeContainer.Create( const InObjectType : TClass;
                                   const InCompareFunc  : TCompareFunc );
begin
    inherited Create(InObjectType);

    FCompareFunc := InCompareFunc;

    Current      := nil;
    Marker       := nil;
end;

{-------------------------------------------------------------------------}

destructor TNodeContainer.Destroy;
begin
    inherited Destroy;
end;

{-------------------------------------------------------------------------}

procedure TNodeContainer.DeleteCurrent;
begin
    ReleaseObj(DetachCurrent);
end;

{-------------------------------------------------------------------------}

procedure TNodeContainer.SetMarker;
begin
    Marker := Current;
end;

{-------------------------------------------------------------------------}

procedure TNodeContainer.GotoMarker;
begin
    if Marker = nil then
        RaiseMarkerNotSet;

    Current := Marker;
end;

{-------------------------------------------------------------------------}

function TNodeContainer.IsOnMarker : Boolean;
begin
    Result := (Marker = Current);
end;

{-------------------------------------------------------------------------}

function TNodeContainer.GetCurrentObj : TObject;
begin
    if Current = nil then
        RaiseContainerIsEmpty;

    Result := Current.Obj;
end;

{-------------------------------------------------------------------------}

procedure TNodeContainer.SetCurrentObj(const Obj : TObject);
begin
    if Current = nil then
        RaiseContainerIsEmpty;

    ReleaseObj(Current.Obj);

    Current.Obj := Obj;
end;

{-------------------------------------------------------------------------}

procedure TNodeContainer.GotoNth(const Index : LongInt);
begin
    GetNthObj(Index);
end;
{-------------------------------------------------------------------------}

procedure ReadContainer( const Obj    : TObject;
                         const Reader : TObjectReader );
var
    Container : TContainer;
begin
    Container := Obj as TContainer;

    Container.FObjectType  := GetRegisteredClass(Reader.ReadString);

    Container.FCompareFunc := GetCompareFunc(Reader.ReadString);

    { ownership is always true when reading from stream }
    Container.FOwnership   := True;

    { the derived container is responsible to update the FSize data member when
      the objects are read. FSize must be cleared to 0. }
    Container.FSize        := 0;
end;

{-------------------------------------------------------------------------}

procedure WriteContainer( const Obj    : TObject;
                          const Writer : TObjectWriter );
var
    Container : TContainer;
begin
    Container := Obj as TContainer;

    Writer.WriteString( Container.FObjectType.ClassName );

    Writer.WriteString( GetCompareFuncName(Container.FCompareFunc) );

    { the derived container is responsible for the FSize data member }
end;

{--------------------------------------------------------------------------}

procedure ReadNodeContainer( const Obj    : TObject;
                             const Reader : TObjectReader );
var
    NodeContainer : TNodeContainer;
    Size          : LongInt;
    Index         : LongInt;
    Item          : TObject;

begin
    NodeContainer := Obj as TNodeContainer;

    ReadContainer(Obj, Reader );

    Size          := Reader.ReadInteger;

    NodeContainer.Current  := nil;
    NodeContainer.Marker   := nil;

    for Index := 1 to Size do
    begin
        Item := Reader.ReadChildObject;

        NodeContainer.Insert(Item);
    end;
end;

{--------------------------------------------------------------------------}

procedure WriteNodeContainer( const Obj    : TObject;
                              const Writer : TObjectWriter );
var
    NodeContainer : TNodeContainer;
begin
    NodeContainer := Obj as TNodeContainer;

    WriteContainer(NodeContainer, Writer);

    Writer.WriteInteger(NodeContainer.Size);

    if NodeContainer.GotoFirst then
    repeat

        Writer.WriteChildObject( NodeContainer.CurrentObj );

    until not NodeContainer.GotoNext;
end;

{--------------------------------------------------------------------------}

end.
