{---------------------------------------------------------------------------
|
|  Module: OdbCompr.Pas
|
|  Description: In memory compression routines which may be used with the
|               Spider Object Database (http://www.intervalsoftware.com)
|
|  RLE Compression routines taken from SWAG Pascal library
|  (http://www.gdsoft.com/swag/swag.html), contributed by Danny Heijl
|  (Danny.Heijl@cevi.be). Described as LZRW1/KH (some sort of run-length
|  encoding), not as good as LZH (Huffman codes) but very fast. Since these
|  routine are taken from SWAG, this unit may be restributed freely in
|  source form.
|
|  History: Jun 08, 1998. MB first version
|           Feb 13, 2001. MB unit tests
|
|---------------------------------------------------------------------------}
unit OdbCompr;     

{$A+} { Word align }
{$R-} { Range check off }
{$Q-} { overflow check off }

interface

uses
    {$ifdef __Testing}
    AutoTest, { for TSpiderOdbTestcase }
    {$endif}

    SysUtils,  { for Exception }
    Classes;   { for TMemoryStream }

{ compression routines using run length encoding. Emphasis is on speed
  rather than compression. }
procedure CompressStreamRLE( const Stream : TMemoryStream );
procedure UnCompressStreamRLE( const Stream : TMemoryStream );

type

EStreamNotCompressed = class(Exception);


{$ifdef __Testing}

type

TOdbComprUnitTest = class( TSpiderOdbTestCase )
published
     procedure TestCompressUnCompress;
end;

{$endif}

{--------------------------------------------------------------------}

implementation

{$ifdef __Testing}
uses
    OdbMisc,  { for CreateStr }
    ObjPerst; { for StreamsEqual }
{$endif}


{---------------------------------------------------------------------------}

{ RLE compression section }

{$ifdef Win32}
type Int16 = SmallInt;
{$else}
type Int16 = Integer;
{$endif}

{---------------------------------------------------------------------------}

const

CFlagCopied     = $80;
CFlagCompress   = $40;

CWorkBufferSize = 60000;
CRleOverhead    = 16;

{---------------------------------------------------------------------------}

type

TByteArray    = array[0..65534] of Byte;
TpByteArray   = ^TByteArray;

THashTable    = array[0..4095] of Int16;
TpHashTable   = ^THashTable;

{---------------------------------------------------------------------------}

{ Check if this string has already been seen in the current 4 KB window }
function GetMatch( const Source     : TpByteArray;
                   const X          : Word;
                   const SourceSize : Word;
                   const pHashTable : TpHashTable;
                   var   Size       : Word;
                   var   Pos        : Word ) : Boolean;
var
    HashValue : Word;
    TempHash  : Int16;
begin
    HashValue := (40543 * ((((Source^[X] shl 4) xor Source^[X + 1]) shl 4) xor
                 Source^[X + 2]) shr 4) and $0fff;

    Result    := False;

    TempHash  := pHashTable^[HashValue];
    if (TempHash <> -1) and (X - TempHash < 4096) then
    begin
        Pos  := TempHash;
        Size := 0;
        while ((Size < 18) and
              (Source^[X + Size] = Source^[Pos + Size]) and
              (X + Size < SourceSize)) do
        begin
            Inc(Size);
        end;
        Result := (Size >= 3)
    end;

    pHashTable^[HashValue] := X;
end;

{---------------------------------------------------------------------------}

{ compress a buffer of max CBufferMaxSize }
function CompressBufferRLE(
                 const Source      : TpByteArray;
                 const Dest        : TpByteArray;
                 const SourceSize  : Word;
                 const pHashTable  : TpHashTable) : Word;
var
    Bit, Command, Size : Word;
    Key                : Word;
    X, Y, Z, Pos       : Word;
begin
    FillChar(pHashTable^, SizeOf(THashTable), $ff);
    Dest^[0] := CFlagCompress;
    X := 0;
    Y := 3;
    Z := 1;
    Bit := 0;
    Command := 0;
    while (X < SourceSize) and (Y <= SourceSize) do
    begin
        if (Bit > 15) then
        begin
            Dest^[Z] := Hi(Command);
            Dest^[Z + 1] := Lo(Command);
            Z := Y;
            Bit := 0;
            Inc(Y, 2)
        end;
        Size := 1;
        while ((Source^[X] = Source^[X + Size]) and
              (Size < $fff) and
              (X + Size < SourceSize)) do
        begin
            Inc(Size);
        end;

        if (Size >= 16) then
        begin
            Dest^[Y] := 0;
            Dest^[Y + 1] := Hi(Size - 16);
            Dest^[Y + 2] := Lo(Size - 16);
            Dest^[Y + 3] := Source^[X];
            Inc(Y, 4);
            Inc(X, Size);
            Command := (Command shl 1) + 1;
        end
        else
        begin { not Size >= 16 }
            if (GetMatch(Source, X, SourceSize, pHashTable, Size, Pos)) then
            begin
                Key := ((X - Pos) shl 4) + (Size - 3);
                Dest^[Y] := Hi(Key);
                Dest^[Y + 1] := Lo(Key);
                Inc(Y, 2);
                Inc(X, Size);
               Command := (Command shl 1) + 1
            end
            else
            begin
                Dest^[Y] := Source^[X];
                Inc(Y);
                Inc(X);
                Command := Command shl 1
            end;
        end; { Size <= 16 }
        Inc(Bit);
    end; { while X < SourceSize ... }

    Command := Command shl (16 - Bit);
    Dest^[Z] := Hi(Command);
    Dest^[Z + 1] := Lo(Command);
    if (Y > SourceSize) then
    begin
        Move(Source^[0], Dest^[1], SourceSize);
        Dest^[0] := CFlagCopied;
        Y := Succ(SourceSize)
    end;

    Result := Y
end;

{---------------------------------------------------------------------------}

{ decompress a buffer }
function UnCompressBufferRLE(
                   const Source     : TpByteArray;
                   const Dest       : TpByteArray;
                   const SourceSize : Word ) : Word;
var
    X, Y, Pos        : Word;
    Command, Size, K : Word;
    Bit              : Byte;
    SaveY            : Word;
begin
    if (SourceSize <= 1) then
    begin { correction of a bug found by Dominique Willems <Domus@compuserve.com>}
        Result := 0;
        Exit;
    end;

    SaveY := 0; { prevent compiler warning }

    if (Source^[0] = CFlagCopied) then
    begin
        for Y := 1 to Pred(SourceSize) do
        begin
            Dest^[Pred(Y)] := Source^[Y];
            SaveY := Y;
        end;
        Y := SaveY;
    end
    else
    begin
        Y := 0;
        X := 3;
        Command := (Source^[1] shl 8) + Source^[2];
        Bit := 16;
        while (X < SourceSize) do
        begin
            if (Bit = 0) then
            begin
                Command := (Source^[X] shl 8) + Source^[X + 1];
                Bit := 16;
                Inc(X, 2)
            end;
            if ((Command and $8000) = 0) then
            begin
                Dest^[Y] := Source^[X];
                Inc(X);
                Inc(Y)
            end
            else
            begin { Command and $8000 }
                Pos := ((Source^[X] shl 4) + (Source^[X + 1] shr 4));
                if (Pos = 0) then
                begin
                    Size := (Source^[X + 1] shl 8) + Source^[X + 2] + 15;
                    for K := 0 to Size do
                        Dest^[Y + K] := Source^[X + 3];
                    Inc(X, 4);
                    Inc(Y, Size + 1)
                end
                else
                begin { pos = 0 }
                    Size := (Source^[X + 1] and $0f) + 2;
                    for K := 0 to Size do
                        Dest^[Y + K] := Dest^[Y - Pos + K];
                    Inc(X, 2);
                    Inc(Y, Size + 1)
                end; { pos = 0 }
            end; { Command and $8000 }
            Command := Command shl 1;
            Dec(Bit);
        end; { while X < SourceSize }
    end;
    Result := Y;
end;

{---------------------------------------------------------------------------}

const

CCompressionSignature = $EFFFFFFE;
CRLESignature         = $70000000;

{---------------------------------------------------------------------------}

{ write a compression signature to Stream }
procedure WriteCompressionSignature( const Stream    : TStream;
                                     const Signature : LongInt );
var
    TempLong : LongInt;
begin
    TempLong := LongInt(CCompressionSignature);
    Stream.Write(TempLong, SizeOf(TempLong));
    TempLong := Signature;
    Stream.Write(TempLong, SizeOf(TempLong));
end;

{---------------------------------------------------------------------------}

{ read the compression signature from Stream }
procedure ReadCompressionSignature( const Stream    : TStream;
                                    const Signature : LongInt );
var
    Temp1 : LongInt;
    Temp2 : LongInt;
begin
    Stream.Read(Temp1, SizeOf(Temp1));
    Stream.Read(Temp2, SizeOf(Temp2));

    if (Temp1 <> LongInt(CCompressionSignature)) or
       (Temp2 <> Signature) then
        raise EStreamNotCompressed.Create(EStreamNotCompressed.ClassName);
end;

{---------------------------------------------------------------------------}

procedure CompressStreamRLE( const Stream : TMemoryStream );
var
    pSourceBuffer : TpByteArray;
    pDestBuffer   : TpByteArray;
    pHashTable    : TpHashTable;

    TempStream    : TMemoryStream;
    SourceSize    : Word;
    DestSize      : Word;

begin
    TempStream    := nil;
    pDestBuffer   := nil;
    pSourceBuffer := nil;
    pHashTable    := nil;
    try
        GetMem(pDestBuffer, CWorkBufferSize);
        GetMem(pSourceBuffer, CWorkBufferSize);
        New(pHashTable);
        TempStream := TMemoryStream.Create;

        TempStream.LoadFromStream(Stream);
        TempStream.Seek(0, soFromBeginning);
        Stream.Clear;
        WriteCompressionSignature(Stream, CRLESignature);

        while TempStream.Position < TempStream.Size do
        begin
            SourceSize := TempStream.Read(pSourceBuffer^, CWorkBufferSize-CRleOverhead);
            DestSize   := CompressBufferRLE( pSourceBuffer, pDestBuffer,
                                             SourceSize, pHashTable );
            Stream.Write(DestSize, SizeOf(DestSize));
            Stream.Write(pDestBuffer^, DestSize);
        end;

    finally
        if TempStream <> nil then
            TempStream.Free;
        if pHashTable <> nil then
            Dispose(pHashTable);
        if pSourceBuffer <> nil then
            FreeMem(pSourceBuffer, CWorkBufferSize);
        if pDestBuffer <> nil then
            FreeMem(pDestBuffer, CWorkBufferSize);
    end;
end;

{---------------------------------------------------------------------------}

procedure UnCompressStreamRLE( const Stream : TMemoryStream );
var
    pSourceBuffer : TpByteArray;
    pDestBuffer   : TpByteArray;
    TempStream    : TMemoryStream;
    SourceSize    : Word;
    DestSize      : Word;

begin
    TempStream    := nil;
    pDestBuffer   := nil;
    pSourceBuffer := nil;
    try
        GetMem(pDestBuffer, CWorkBufferSize);
        GetMem(pSourceBuffer, CWorkBufferSize);
        TempStream := TMemoryStream.Create;

        TempStream.LoadFromStream(Stream);
        TempStream.Seek(0, soFromBeginning);
        Stream.Clear;
        ReadCompressionSignature(TempStream, CRLESignature);

        while TempStream.Read(SourceSize, SizeOf(SourceSize)) > 0 do
        begin
           SourceSize := TempStream.Read(pSourceBuffer^, SourceSize);
           DestSize   := UnCompressBufferRLE(pSourceBuffer, pDestBuffer, SourceSize);
           Stream.Write(pDestBuffer^, DestSize);
        end;
    finally
        if TempStream <> nil then
            TempStream.Free;
        if pSourceBuffer <> nil then
            FreeMem(pSourceBuffer, CWorkBufferSize);
        if pDestBuffer <> nil then
            FreeMem(pDestBuffer, CWorkBufferSize);
    end;
end;

{---------------------------------------------------------------------------}

{$ifdef __Testing}

procedure FillStream( const Stream  : TStream;
                      const MinSize : LongInt );
var
    Counter      : LongInt;
    WordLength : Integer;
    Word       : String;
begin
    Counter := 0;
    while Stream.Size <= MinSize do
    begin
        WordLength := (Counter mod 10) + 1;
        Word       := OdbMisc.CreateStr( Char(Counter mod 255), WordLength);

        Stream.Write( Word[1], Length(Word) );

        Inc(Counter);
    end;
    
    Stream.Seek(0, soFromBeginning);
end;

procedure TOdbComprUnitTest.TestCompressUncompress;
var
    OriginalStream      : TMemoryStream;
    WorkStream          : TMemoryStream;
    TempByte            : Byte;

    procedure LocalTest( const StreamSize : LongInt );
    begin
        OriginalStream.Clear;
        FillStream( OriginalStream, StreamSize );

        WorkStream.Clear;
        WorkStream.CopyFrom( OriginalStream, OriginalStream.Size );

        CompressStreamRLE( WorkStream );

        UncompressStreamRLE( WorkStream );

        Check( ObjPerst.StreamsEqual( OriginalStream, WorkStream ) );
    end;

begin
    OriginalStream  := TMemoryStream.Create;
    WorkStream      := TMemoryStream.Create;
    try
        { test on empty stream }
        CompressStreamRLE( WorkStream );
        UncompressStreamRLE( WorkStream );
        Check( WorkStream.Size = 0 );

        { check with a single byte }
        TempByte := 1;
        WorkStream.Clear;
        WorkStream.Write( TempByte, SizeOf(TempByte));
        CompressStreamRLE( WorkStream );
        UncompressStreamRLE( WorkStream );
        Check( WorkStream.Size = 1 );
        WorkStream.Seek(0, soFromBeginning);
        WorkStream.Read( TempByte, SizeOf(TempByte) );
        Check( TempByte = 1 );

        { check with a single NULL byte }
        TempByte := 0;
        WorkStream.Clear;
        WorkStream.Write( TempByte, SizeOf(TempByte));
        CompressStreamRLE( WorkStream );
        UncompressStreamRLE( WorkStream );
        Check( WorkStream.Size = 1 );
        WorkStream.Seek(0, soFromBeginning);
        WorkStream.Read( TempByte, SizeOf(TempByte) );
        Check( TempByte = 0 );

        LocalTest(2);
        LocalTest(10);
        LocalTest(102);
        LocalTest(1045);
        LocalTest(10200);
        LocalTest(104001);
        LocalTest(1000023);

    finally
        OriginalStream.Free;
        WorkStream.Free;
    end;
end;

{$endif}

{---------------------------------------------------------------------------}

end.

