{----------------------------------------------------------------------
|
| Library: Spider Object Database
|
| Module: OdbDBISM.Pas
|
| Description: DBISAM object database driver class.
|
| History: Jan 03, 1998. Michel Brazeau, first version
|          Oct 05, 2000. MB adapt for DBISAM version 2. Support damSQL
|                        access method.
|          Mar 01, 2001. MB override introspection methods and testcase
|
|---------------------------------------------------------------------}
unit OdbDbIsm;

{$I Odb.Inc}

interface

uses
    {$ifdef __Testing}
    DriverTest, { for TOdbDriverTest }
    {$endif}

    Classes,  { for TMemoryStream }
    SysUtils, { for Exception, StrPas }
    ObjList,  { for TUnOrderedList }
    DbIsamTb, { for TDBISAMDataset, TDBISAMTable, TDBISAMQuery }
    Odb,      { for TObjectFile, TOdbDriver }
    OdbIndex, { for TOdbIndex }
    OdbField; { for TOdbField }

type

TOdbDbIsamTable = class(TDbIsamTable)
public
    { for DbIsam tables, the primary index name must be ''. PrimaryIndexName
      is the index name as defined by the related object file. }
    PrimaryIndexName : String;
end;
{--------------------------------------------------------------------}

TDBISAMOdbDriver = class(TOdbDriver)
protected

    FDatabase      : TDBISAMDatabase;

public

    constructor Create(const InDatabaseName : String); override;
    destructor Destroy; override;

    procedure InternalStartTransaction; override;
    procedure InternalCommit; override;
    procedure InternalRollback; override;

    procedure InternalOpenDatabase(
                            const ServerName : String;
                            const UserName   : String;
                            const Password   : String
                          ); override;

    procedure InternalCloseDatabase; override;

    function InternalOpenDataset( const ObjectDataset : TObjectDataset ) : TOdbHandle; override;

    function InternalOpenExclusive( const ObjectFile : TObjectFile ) : TOdbHandle; override;

    procedure InternalCloseDataset( const Handle : TOdbHandle); override;

    procedure InternalDeleteDataset( const ObjectDataset : TObjectDataset ); override;

    procedure InternalCreateDataset( const ObjectDataset : TObjectDataset ); override;

    procedure InternalInsert(
                      const Handle  : TOdbHandle;
                      const Fields  : TUnOrderedList;
                      const Stream  : TMemoryStream ); override;

    procedure InternalUpdate(
                      const Handle     : TOdbHandle;
                      const Fields     : TUnOrderedList;
                      const Stream     : TMemoryStream ); override;

    procedure InternalDelete( const Handle : TOdbHandle ); override;

    function InternalCreateObjectUpdate : TOdbHandle; override;
    procedure InternalDestroyObjectUpdate( const Handle : TOdbHandle ); override;

    procedure InternalInsertSQL(
                      const Handle      : TOdbHandle;
                      const Stream      : TMemoryStream;
                      const ORMapping   : TORMapping;
                      const DatasetName : String ); override;

    procedure InternalUpdateSQL(
                      const Handle      : TOdbHandle;
                      const Stream      : TMemoryStream;
                      const ORMapping   : TORMapping;
                      const DatasetName : String ); override;

    procedure InternalDeleteSQL( const Handle      : TOdbHandle;
                                 const ORMapping   : TORMapping;
                                 const DatasetName : String ); override;

    function InternalGetIndexName( const Handle : TOdbHandle ) : String; override;

    procedure InternalSetIndexName(
                            const Handle    : TOdbHandle;
                            const IndexName : String ); override;

    procedure InternalGetCurrent(
                          const Handle  : TOdbHandle;
                          const Stream  : TMemoryStream ); override;

    function InternalLockCurrent( const Handle : TOdbHandle ) : Boolean; override;

    procedure InternalUnLockCurrent( const Handle : TOdbHandle ); override;

    function InternalFindNext( const Handle : TOdbHandle ) : Boolean; override;

    function InternalFindPrior( const Handle : TOdbHandle ) : Boolean; override;

    function InternalFindFirst( const Handle : TOdbHandle ) : Boolean; override;

    function InternalFindLast( const Handle : TOdbHandle ) : Boolean; override;

    function InternalFindExact(
                        const Handle : TOdbHandle;
                        const Fields : TUnOrderedList;
                        const Index  : TOdbIndex ) : Boolean; override;

    procedure InternalFindNearest(
                          const Handle : TOdbHandle;
                          const Fields : TUnOrderedList;
                          const Index  : TOdbIndex ); override;

    function InternalGetBookmark( const Handle : TOdbHandle ) : Pointer; override;
    procedure InternalGotoBookmark( const Handle   : TOdbHandle;
                                    const Bookmark : Pointer ); override;
    procedure InternalFreeBookmark( const Handle   : TOdbHandle;
                                    const Bookmark : Pointer ); override;

    function InternalGetObjectCount( const Handle : TOdbHandle ) : LongInt; override;

    function InternalIsEmpty( const Handle : TOdbHandle ) : Boolean; override;

    function GetSupportedFieldTypes : TOdbFieldTypes; override;

    function CheckDuplicateInsert : Boolean; override;
    function CheckDuplicateUpdate : Boolean; override;

    function GetDBISAMDataset( const Handle : TOdbHandle ) : TDBISAMDataset;

    function GetDataAccessMethods : TDataAccessMethodSet; override;


    function AutoIncSupported              : Boolean; override;
    function IntegerIndexSupported         : Boolean; override;
    function FloatIndexSupported           : Boolean; override;
    function CaseInsensitiveIndexSupported : Boolean; override;
    function DescendingIndexSupported      : Boolean; override;
    function TransactionsSupported         : Boolean; override;
    function FileLockingSupported          : Boolean; override;
    function ObjectLockingSupported        : Boolean; override;
    function SingleIndexOnly               : Boolean; override;


protected
    function DbIsamOpenFile( const ObjectFile : TObjectFile;
                             const Exclusive  : Boolean ) : TOdbHandle;

    function DbIsamOpenQuery( const ObjectQuery : TObjectQuery ) : TOdbHandle;

end;                                             


{$ifdef __Testing}

type

TDbIsamOdbDriverTest = class( TOdbDriverTest )
protected
     procedure GetParams( var DriverClass                   : TOdbDriverClass;
                          var DatabaseName                  : String;
                          var ServerName                    : String;
                          var UserName                      : String;
                          var Password                      : String
                        ); override;
end;

{$endif}

{--------------------------------------------------------------------}

implementation

uses
    Db,       { for ftString, ftInteger, ftFloat }
    Bde,      { DBIERR_XXXX }
    DbIsamEn, { TDataFile }
    ObjPerst,
    ObjBuckt, { for TStringCombo }
    OdbMisc;  { for RaiseOdbException }

const

CObjectFieldName = 'Object';

{--------------------------------------------------------------------}

var

{ last database driver instance that was created. }
LastInstanceNo   : LongInt;

{--------------------------------------------------------------------}

procedure DbiCheck(Status : Integer);
begin
    DbIsamTb.Check(Status);
end;

{--------------------------------------------------------------------}

constructor TDBISAMOdbDriver.Create(const InDatabaseName : String);
begin
    FDatabase := TDBISAMDatabase.Create(nil);

    Inc(LastInstanceNo);

    FDatabase.Directory    := InDatabaseName;
    FDatabase.Databasename := 'Odb' + IntToStr(LastInstanceNo);

    inherited Create('Odb');
end;

{--------------------------------------------------------------------}

destructor TDBISAMOdbDriver.Destroy;
begin
    { invoke inherited first, to close the database if it is opened before
      invoking FDatabase.Free }
    inherited Destroy;

    FDatabase.Free;
end;

{--------------------------------------------------------------------}

procedure TDBISAMOdbDriver.InternalStartTransaction;
begin
    FDatabase.StartTransaction;
end;

{--------------------------------------------------------------------}

procedure TDBISAMOdbDriver.InternalCommit;
begin
    FDatabase.Commit;
end;

{--------------------------------------------------------------------}

procedure TDBISAMOdbDriver.InternalRollback;
begin
    FDatabase.Rollback;
end;

{--------------------------------------------------------------------}

procedure TDBISAMOdbDriver.InternalOpenDatabase(
                        const ServerName : String;
                        const UserName   : String;
                        const Password   : String
                      );
begin
    FDatabase.KeepConnection := True;

    FDatabase.Open;
end;

{--------------------------------------------------------------------}

procedure TDBISAMOdbDriver.InternalCloseDatabase;
begin
    FDatabase.Close;
end;

{--------------------------------------------------------------------}

function TDBISAMOdbDriver.DbIsamOpenFile( const ObjectFile : TObjectFile;
                                          const Exclusive  : Boolean ) : TOdbHandle;
var
    Table : TOdbDbIsamTable;
    Index : TOdbIndex;

begin
    Table := TOdbDbIsamTable.Create(nil);
    try
        Table.DatabaseName := FDatabase.DatabaseName;
        Table.TableName    := ObjectFile.DatasetName;
        Table.Exclusive    := Exclusive;
        Table.Open;

        Result := NewHandle(Table);
    except
        Table.Free;
        raise;
    end;

    if ObjectFile.Indices.GotoFirst then
    repeat
        Index := (ObjectFile as TObjectFile).Indices.CurrentObj as TOdbIndex;

        if iaPrimary in Index.Attributes then
        begin
            Table.PrimaryIndexName := Index.Name;
            break;
        end;
    until not ObjectFile.Indices.GotoNext;
end;

{--------------------------------------------------------------------}

function TDBISAMOdbDriver.DbIsamOpenQuery(
                           const ObjectQuery : TObjectQuery ) : TOdbHandle;
var
    Query     : TDbIsamQuery;

begin
    Query     := nil;
    try
        Query     := TDbIsamQuery.Create(nil);

        Query.DatabaseName := FDatabase.DatabaseName;
        Query.SQL.Assign( ObjectQuery.Select );
        Query.Open;

        Result := NewHandle(Query);
    except
        Query.Free;
        raise;
    end;
end;

{--------------------------------------------------------------------}

function TDBISAMOdbDriver.InternalOpenDataset(
                                      const ObjectDataset : TObjectDataset
                                      ) : TOdbHandle;
begin
    if ObjectDataset is TObjectFile then
        Result := DbIsamOpenFile(TObjectFile(ObjectDataset), False)
    else if ObjectDataset is TObjectQuery then
    begin
        if TObjectQuery(ObjectDataset).Select.Count = 0 then
            RaiseOdbException( EOdbSelectStatementNotSpecified,
                               ObjectDataset.DatasetName );

        Result := DbIsamOpenQuery( TObjectQuery(ObjectDataset) );
    end
    else
    begin
        Result := 0; { prevent compiler warning }
        RaiseOdbException(EOdbError, 'Internal error ' + ObjectDataset.DatasetName);
    end;
end;
{--------------------------------------------------------------------}

function TDBISAMOdbDriver.InternalOpenExclusive(
                                             const ObjectFile : TObjectFile
                                            ) : TOdbHandle;
begin
    Result := DbIsamOpenFile(ObjectFile, True);
end;

{--------------------------------------------------------------------}

procedure TDBISAMOdbDriver.InternalCloseDataset( const Handle : TOdbHandle);
var
    Dataset : TDBISAMDataset;
begin
    Dataset := GetDBISAMDataset(Handle);

    Dataset.Close;

    ReleaseHandle(Handle);
end;

{--------------------------------------------------------------------}

function TDBISAMOdbDriver.InternalCreateObjectUpdate : TOdbHandle;
var
    EditQuery : TDbIsamQuery;

begin
    EditQuery := nil;
    try
        EditQuery := TDbIsamQuery.Create(nil);
        EditQuery.DatabaseName := FDatabase.DatabaseName;

        Result := NewHandle(EditQuery);
    except
        EditQuery.Free;
        raise;
    end;
end;

{--------------------------------------------------------------------}

procedure TDBISAMOdbDriver.InternalDestroyObjectUpdate( const Handle : TOdbHandle );
begin
    ReleaseHandle(Handle);
end;

{--------------------------------------------------------------------}

procedure TDBISAMOdbDriver.InternalDeleteDataset(
                                     const ObjectDataset : TObjectDataset );

var
    Table : TOdbDbIsamTable;
begin
    Table := TOdbDbIsamTable.Create(nil);
    try
        Table.DatabaseName := FDatabase.DatabaseName;
        Table.TableName    := ObjectDataset.DatasetName;
        Table.DeleteTable;
    finally
        Table.Free;
    end;
end;

{--------------------------------------------------------------------}

{ add Fields to Table }
procedure AddFields( const Fields  : TUnOrderedList;
                     const Table   : TOdbDbIsamTable );
var
    Field   : TOdbField;
begin
    if Fields.GotoFirst then
    repeat
        Field := Fields.CurrentObj as TOdbField;

        case Field.FieldType of
            oftString  : Table.FieldDefs.Add(
                                 Field.Name,
                                 Db.ftString, Field.Size, False );
            oftInteger : Table.FieldDefs.Add(
                                 Field.Name,
                                 Db.ftInteger, 0, False );
            oftFloat   : Table.FieldDefs.Add(
                                 Field.Name,
                                 Db.ftFloat, 0, False );
            oftAutoInc : Table.FieldDefs.Add(
                                 Field.Name,
                                 Db.ftAutoInc, 0, False );

        end;
    until not Fields.GotoNext;

    Table.FieldDefs.Add(CObjectFieldName, ftBlob, 0, False);
end;

{--------------------------------------------------------------------}

{ convert Spider Odb index attributes to VCL index options }
function IndexAttributesToOptions(
            const Attributes : TOdbIndexAttributes ) : TIndexOptions;
begin
    Result := [];

    if iaUnique in Attributes then
        Result := Result + [ixUnique];

    if iaCaseInsensitive in Attributes then
        Result := Result + [ixCaseInsensitive];

    if iaDescending in Attributes then
        Result := Result + [ixDescending];

    if iaPrimary in Attributes then
        Result := Result + [ixPrimary];
end;

{--------------------------------------------------------------------}

procedure AddIndices( const Indices          : TUnOrderedList;
                      const Table            : TOdbDbIsamTable );
var
    Index        : TOdbIndex;
    IndexOptions : TIndexOptions;
    FieldNames   : String;
    IndexName    : String;

begin
    if Indices.GotoFirst then
    repeat
        Index        := Indices.CurrentObj as TOdbIndex;
        FieldNames   := Index.FieldNamesAsString(';');
        IndexOptions := IndexAttributesToOptions(Index.Attributes);

        IndexName := Index.Name;

        if (ixPrimary in IndexOptions) then
            IndexName := ''; {  primary index must be = '' }

        Table.IndexDefs.Add( IndexName,
                             FieldNames,
                             IndexOptions );


    until not Indices.GotoNext;
end;

{--------------------------------------------------------------------}

procedure TDBISAMOdbDriver.InternalCreateDataset(
                          const ObjectDataset : TObjectDataset
                        );
var
    Table : TOdbDbIsamTable;

begin
    Table := TOdbDbIsamTable.Create(nil);
    try
        Table.DatabaseName := FDatabase.DatabaseName;
        Table.TableName    := ObjectDataset.DatasetName;

        Table.FieldDefs.Clear;
        Table.IndexDefs.Clear;

        AddFields(ObjectDataset.Fields, Table);
        AddIndices(ObjectDataset.Indices, Table);
        Table.CreateTable;

    finally
        Table.Free;
    end;
end;

{--------------------------------------------------------------------}

{ set the field values in Table given a list of TOdbField's }
procedure SetDatasetFieldValues( const Fields  : TUnOrderedList;
                                 const Dataset   : TDBISAMDataset );
var
    Field : TOdbField;
begin
    if Fields.GotoFirst then
    repeat
        Field := Fields.CurrentObj as TOdbField;

        case Field.FieldType of
            oftString  : Dataset.FieldByName(Field.Name).AsString  := Field.AsString;
            oftInteger : Dataset.FieldByName(Field.Name).AsInteger := Field.AsInteger;
            oftFloat   : Dataset.FieldByName(Field.Name).AsFloat   := Field.AsFloat;
        end;
    until not Fields.GotoNext;
end;

{--------------------------------------------------------------------}

{ set the field values in Table for the fields belonging to Index }
procedure SetTableFieldValuesOnIndex( const Fields  : TUnOrderedList;
                                      const Index   : TOdbIndex;
                                      const Table   : TOdbDbIsamTable );
var
    Field       : TOdbField;
    StringCombo : TStringCombo;
begin
    if Index.FieldNames.GotoFirst then
    repeat
        StringCombo := Index.FieldNames.CurrentObj as TStringCombo;

        Field := GetFieldInList( StringCombo.Str, Fields );

        case Field.FieldType of
            oftString  : Table.FieldByName(Field.Name).AsString := Field.AsString;
            oftInteger,
            oftAutoInc : Table.FieldByName(Field.Name).AsInteger := Field.AsInteger;
            oftFloat   : Table.FieldByName(Field.Name).AsFloat := Field.AsFloat;
        end;
    until not Index.FieldNames.GotoNext;
end;

{--------------------------------------------------------------------}

{ Set the value of the autoincrement fields in Fields from the values
  contained in Table. }
procedure SetOdbAutoIncFieldValues( const Fields  : TUnOrderedList;
                                    const Table   : TOdbDbIsamTable );
var
    Field : TOdbField;
begin
    if Fields.GotoFirst then
    repeat
        Field := Fields.CurrentObj as TOdbField;

        case Field.FieldType of
            oftAutoInc : Field.AsInteger := Table.FieldByName(Field.Name).AsInteger;
        end;
    until not Fields.GotoNext;
end;

{--------------------------------------------------------------------}

procedure InsertInDataset( const Dataset : TDBISAMDataset;
                           const Fields  : TUnOrderedList;
                           const Stream  : TMemoryStream );
var
    Field   : TBlobField;
begin
    Dataset.Insert;
    try
        SetDatasetFieldValues(Fields, Dataset);

        Field := (Dataset.FieldByName(CObjectFieldName) as TBlobField);

        Field.LoadFromStream(Stream);

        Dataset.Post;

        { Update any autoincrement field in Fields. }
        if Dataset is TOdbDbIsamTable then
             SetOdbAutoIncFieldValues(Fields, TOdbDbIsamTable(Dataset));
    except
        Dataset.Cancel;
        raise;
    end;
end;

{--------------------------------------------------------------------}

procedure TDBISAMOdbDriver.InternalInsert(
                  const Handle  : TOdbHandle;
                  const Fields  : TUnOrderedList;
                  const Stream  : TMemoryStream );

var
    Dataset : TDBISAMDataset;
begin
    Dataset := GetDBISAMDataset(Handle);

    InsertInDataset(Dataset, Fields, Stream);
end;

{--------------------------------------------------------------------}

procedure UpdateCurrentInDataset( const Dataset : TDBISAMDataset;
                                  const Fields  : TUnOrderedList;
                                  const Stream  : TMemoryStream );
begin
     try
        SetDatasetFieldValues(Fields, Dataset);
        (Dataset.FieldByName(CObjectFieldName) as TBlobField).LoadFromStream(Stream);
        Dataset.Post;
    except
        Dataset.Cancel;
        raise;
    end;
end;

{--------------------------------------------------------------------}

procedure TDBISAMOdbDriver.InternalUpdate(
                  const Handle     : TOdbHandle;
                  const Fields     : TUnOrderedList;
                  const Stream     : TMemoryStream );

var
    Dataset      : TDBISAMDataset;
begin
    Dataset       := GetDBISAMDataset(Handle);

    { MB May 21, 1998. Dataset.Edit will attempt to lock the record before
      updating it. If the record has already been locked by the current session,
      it is acceptable to perform an update. A try,except block is used to
      attempt to unlock the object. If the object is locked by another session,
      the second call to Dataset.Edit will raise the exception. }
    try
        Dataset.Edit;
    except
        on Exception do
        begin
            InternalUnLockCurrent(Handle);
            Dataset.Edit;
        end;
    end;

    UpdateCurrentInDataset(Dataset, Fields, Stream);
end;

{--------------------------------------------------------------------}

procedure TDBISAMOdbDriver.InternalDelete( const Handle : TOdbHandle );
var
    Dataset       : TDBISAMDataset;
begin
    Dataset       := GetDBISAMDataset(Handle);

    Dataset.Delete
end;

{--------------------------------------------------------------------}

procedure TDBISAMOdbDriver.InternalInsertSQL(
                  const Handle      : TOdbHandle;
                  const Stream      : TMemoryStream;
                  const ORMapping   : TORMapping;
                  const DatasetName : String );
var
    EditQuery     : TDbIsamQuery;
begin
    EditQuery := GetHandleDataObj(Handle) as TDBISAMQuery;

    BuildInsertStatement( EditQuery.SQL, ORMapping.Fields, DatasetName );
    EditQuery.ExecSQL;

    { add the blob }
    BuildSelectObjectQuery( EditQuery.SQL, ORMapping, DatasetName );
    EditQuery.RequestLive := True;
    EditQuery.Open;
    try
        if EditQuery.Bof and EditQuery.Eof then
            RaiseOdbException(EOdbObjectNotFound, 'TDBISAMOdbDriver.InternalUpdateSQL');

        EditQuery.Edit;

        UpdateCurrentInDataset(EditQuery, ORMapping.Fields, Stream);
    finally
        EditQuery.Close;
    end;
end;

{--------------------------------------------------------------------}

procedure TDBISAMOdbDriver.InternalUpdateSQL(
                  const Handle      : TOdbHandle;
                  const Stream      : TMemoryStream;
                  const ORMapping   : TORMapping;
                  const DatasetName : String );
var
    EditQuery     : TDbIsamQuery;

begin
    { Update the record using the related edit query. The record must
      first be located. }
    EditQuery := GetHandleDataObj(Handle) as TDBISAMQuery;

    BuildSelectObjectQuery( EditQuery.SQL, ORMapping, DatasetName );

    EditQuery.Open;
    try
        if EditQuery.Bof and EditQuery.Eof then
            RaiseOdbException(EOdbObjectNotFound, 'TDBISAMOdbDriver.InternalUpdateSQL');

        EditQuery.Edit;

        UpdateCurrentInDataset(EditQuery, ORMapping.Fields, Stream);
    finally
        EditQuery.Close;
    end;
end;

{--------------------------------------------------------------------}

procedure TDBISAMOdbDriver.InternalDeleteSQL(
                             const Handle      : TOdbHandle;
                             const ORMapping   : TORMapping;
                             const DatasetName : String );
var
    EditQuery     : TDbIsamQuery;

begin
    { Delete the record using the related edit query. }
    EditQuery := GetHandleDataObj(Handle) as TDBISAMQuery;

    BuildDeleteStatement( EditQuery.SQL, ORMapping, DatasetName );

    EditQuery.ExecSQL;
end;

{--------------------------------------------------------------------}

function TDBISAMOdbDriver.InternalGetIndexName( const Handle : TOdbHandle ) : String;
var
    Table : TOdbDbIsamTable;
begin
    Table  := GetDBISAMDataset(Handle) as TOdbDbIsamTable;
    Result := Table.IndexName;

    if Result = '' then
        Result := Table.PrimaryIndexName;
end;

{--------------------------------------------------------------------}

procedure TDBISAMOdbDriver.InternalSetIndexName(
                      const Handle    : TOdbHandle;
                      const IndexName : String );
var
    Table : TOdbDbIsamTable;
begin
    Table := GetDbIsamDataset(Handle) as TOdbDbIsamTable;

    if (IndexName = Table.PrimaryIndexName) then
        Table.IndexName := ''
    else
        Table.IndexName := IndexName;
end;

{--------------------------------------------------------------------}

procedure TDBISAMOdbDriver.InternalGetCurrent(
                      const Handle  : TOdbHandle;
                      const Stream  : TMemoryStream );
var
    Dataset : TDBISAMDataset;
    Field   : TBlobField;
begin
    Dataset := GetDBISAMDataset(Handle) as TDBISAMDataset;

    Field := (Dataset.FieldByName(CObjectFieldName) as TBlobField);

    Field.SaveToStream(Stream);
end;

{--------------------------------------------------------------------}

function TDBISAMOdbDriver.InternalLockCurrent( const Handle : TOdbHandle ) : Boolean;
var
    Table    : TOdbDbIsamTable;
begin
    Table := GetDbIsamDataset(Handle) as TOdbDbIsamTable;

    Table.UpdateCursorPos;

    { MB Mar 10, 2001. Rewrite.  }
    try
        Result := True;
        Table.Handle.UserWriteLockRecord(0);
    except
        Result := False;
    end;
end;


{--------------------------------------------------------------------}

procedure TDBISAMOdbDriver.InternalUnLockCurrent( const Handle : TOdbHandle );
var
    Table : TOdbDbIsamTable;
begin
    Table := GetDbIsamDataset(Handle) as TOdbDbIsamTable;

    Table.UpdateCursorPos;

    { MB May 21, 1998. Don't raise an exception if the current object is
      not locked. This is important as the InternalUpdate method may attempt
      to unlock an object which is not locked. }
    try
        Table.Handle.UserWriteUnlockRecord(0);
    except
    end;
end;


{--------------------------------------------------------------------}

function TDBISAMOdbDriver.InternalFindNext( const Handle : TOdbHandle ) : Boolean;
var
    Dataset : TDBISAMDataset;
begin
    Dataset := GetDBISAMDataset(Handle);
    Dataset.Next;
    Result := not Dataset.EOF;
end;

{--------------------------------------------------------------------}

function TDBISAMOdbDriver.InternalFindPrior( const Handle : TOdbHandle ) : Boolean;
var
    Dataset : TDBISAMDataset;
begin
    Dataset := GetDBISAMDataset(Handle);
    Dataset.Prior;
    Result := not Dataset.BOF;
end;

{--------------------------------------------------------------------}

function TDBISAMOdbDriver.InternalFindFirst( const Handle : TOdbHandle ) : Boolean;
var
    Dataset : TDBISAMDataset;
begin
    Dataset := GetDBISAMDataset(Handle);
    Dataset.First;
    Result := not (Dataset.EOF and Dataset.BOF);
end;

{--------------------------------------------------------------------}

function TDBISAMOdbDriver.InternalFindLast( const Handle : TOdbHandle ) : Boolean;
var
    Dataset : TDBISAMDataset;
begin
    Dataset := GetDBISAMDataset(Handle);
    Dataset.Last;
    Result := not (Dataset.EOF and Dataset.BOF);
end;

{--------------------------------------------------------------------}

function TDBISAMOdbDriver.InternalFindExact(
                    const Handle : TOdbHandle;
                    const Fields : TUnOrderedList;
                    const Index  : TOdbIndex ) : Boolean;
var
    Table : TOdbDbIsamTable;
begin
    Table := GetDBISAMDataset(Handle) as TOdbDbIsamTable;
    Table.SetKey;
    SetTableFieldValuesOnIndex(Fields, Index, Table);
    Result := Table.GotoKey;
end;

{--------------------------------------------------------------------}

procedure TDBISAMOdbDriver.InternalFindNearest(
                      const Handle : TOdbHandle;
                      const Fields : TUnOrderedList;
                      const Index  : TOdbIndex );
var
    Table : TOdbDbIsamTable;
begin
    Table := GetDBISAMDataset(Handle) as TOdbDbIsamTable;
    Table.SetKey;
    SetTableFieldValuesOnIndex(Fields, Index, Table);
    Table.GotoNearest;
end;

{--------------------------------------------------------------------}

function TDBISAMOdbDriver.InternalGetBookmark( const Handle : TOdbHandle ) : Pointer;
var
    Table : TOdbDbIsamTable;
begin
    Table := GetDBISAMDataset(Handle) as TOdbDbIsamTable;

    Result := Table.GetBookmark;
end;

{--------------------------------------------------------------------}

procedure TDBISAMOdbDriver.InternalGotoBookmark( const Handle   : TOdbHandle;
                                              const Bookmark : Pointer );
var
    Table : TOdbDbIsamTable;
begin
    Table := GetDBISAMDataset(Handle) as TOdbDbIsamTable;

    Table.GotoBookmark(Bookmark);
end;

{--------------------------------------------------------------------}

procedure TDBISAMOdbDriver.InternalFreeBookmark( const Handle   : TOdbHandle;
                                              const Bookmark : Pointer );
var
    Table : TOdbDbIsamTable;
begin
    Table := GetDBISAMDataset(Handle) as TOdbDbIsamTable;

    Table.FreeBookmark(Bookmark);
end;


{--------------------------------------------------------------------}

function TDBISAMOdbDriver.InternalGetObjectCount( const Handle : TOdbHandle ) : LongInt;
begin
    Result := (GetDBISAMDataset(Handle) as TOdbDbIsamTable).RecordCount;
end;

{--------------------------------------------------------------------}

function TDBISAMOdbDriver.InternalIsEmpty( const Handle : TOdbHandle ) : Boolean;
var
    Dataset : TDbIsamDataset;
begin
    Dataset := GetDBISAMDataset(Handle);
    Result  := Dataset.BOF and Dataset.EOF;
end;

{--------------------------------------------------------------------}

function TDBISAMOdbDriver.GetDBISAMDataset( const Handle : TOdbHandle ) : TDBISAMDataset;
begin
    Result := GetHandleDataObj(Handle) as TDBISAMDataset;
end;

{--------------------------------------------------------------------}

function TDBISAMOdbDriver.GetDataAccessMethods : TDataAccessMethodSet;
begin
    Result := [damISAM, damSQL];
end;

{--------------------------------------------------------------------}

function TDBISAMOdbDriver.GetSupportedFieldTypes : TOdbFieldTypes;
begin
    Result := [oftString, oftInteger, oftFloat, oftAutoInc];
end;

{--------------------------------------------------------------------}

function TDBISAMOdbDriver.CheckDuplicateInsert : Boolean;
begin
    Result := False;
end;

{--------------------------------------------------------------------}

function TDBISAMOdbDriver.CheckDuplicateUpdate : Boolean;
begin
    Result := False;
end;

{--------------------------------------------------------------------------}

function TDBISAMOdbDriver.AutoIncSupported              : Boolean;
begin
    Result := True;
end;

{--------------------------------------------------------------------------}

function TDBISAMOdbDriver.IntegerIndexSupported         : Boolean;
begin
    Result := True;
end;

{--------------------------------------------------------------------------}

function TDBISAMOdbDriver.FloatIndexSupported           : Boolean;
begin
    Result := True;
end;

{--------------------------------------------------------------------------}

function TDBISAMOdbDriver.CaseInsensitiveIndexSupported : Boolean;
begin
    Result := True;
end;

{--------------------------------------------------------------------------}

function TDBISAMOdbDriver.DescendingIndexSupported      : Boolean;
begin
    Result := True;
end;

{--------------------------------------------------------------------------}

function TDBISAMOdbDriver.TransactionsSupported         : Boolean;
begin
    Result := True;
end;

{--------------------------------------------------------------------------}

function TDBISAMOdbDriver.FileLockingSupported          : Boolean;
begin
    Result := True;
end;

{--------------------------------------------------------------------------}

function TDBISAMOdbDriver.ObjectLockingSupported        : Boolean;
begin
    Result := True;
end;

{--------------------------------------------------------------------------}

function TDBISAMOdbDriver.SingleIndexOnly               : Boolean;
begin
    Result := False;
end;

{--------------------------------------------------------------------}

{$ifdef __Testing}

procedure TDbIsamOdbDriverTest.GetParams(
                    var DriverClass                   : TOdbDriverClass;
                    var DatabaseName                  : String;
                    var ServerName                    : String;
                    var UserName                      : String;
                    var Password                      : String
                  );
begin
    DriverClass  := TDBISAMOdbDriver;
    DatabaseName := Self.TestDbPath;
    ServerName   := '';
    UserName     := '';
    Password     := '';
end;

{$endif}

{--------------------------------------------------------------------}


initialization
    LastInstanceNo := 0;
    
    RegisterOdbDriver(TDBISAMOdbDriver, 'DBISAM');
end.
